/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "math.pmh"

namespace fe
{

void DAGNode::validate(void)
{
	//* verify that each parent references this node exactly once
	List< hp<Connection> >::Iterator itConnection(m_parentConnections);
	hp<Connection> hpConnection;
	while((hpConnection= *itConnection++).isValid())
	{
		List< sp<Connection> >::Iterator itConnection(
				hpConnection->parent()->m_childConnections);
		sp<Connection> spConnection;
		U32 count=0;
		while((spConnection= *itConnection++).isValid())
		{
			if(spConnection->child().raw()==this)
			{
				count++;
			}
		}
		if(count!=1)
		{
			feX("DAGNode::validate","found %d parent references",count);
		}
	}
}

BWORD DAGNode::attachTo(sp<DAGNode> spNode,
	String a_localConnector,String a_remoteConnector)
{
	FE_DAG_VALIDATE();
	ListCore::Context context;
	if(spNode->m_childConnections.searchForElement(context,spNode))
	{
		feLog("DAGNode::attachTo node already attached\n");
		return FALSE;
	}

	sp<Connection> spConnection(new Connection(sp<DAGNode>(this),spNode,
			a_localConnector,a_remoteConnector));

	spNode->m_childConnections.append(spConnection);
	spNode->m_childMap[a_remoteConnector]=spConnection;

	m_parentConnections.append(spConnection);
	m_parentMap[a_localConnector]=spConnection;

	FE_DAG_VALIDATE();
	return TRUE;
}

BWORD DAGNode::detachFrom(sp<DAGNode> spNode)
{
	FE_DAG_VALIDATE();

	ListCore::Context context;
	sp<Connection>* pspConnection;
	if(pspConnection=spNode->m_childConnections.searchForElement(
			context,sp<DAGNode>(this)))
	{
		const String childConnector=(*pspConnection)->childConnector();
		spNode->m_childConnections.remove(sp<DAGNode>(this));
		spNode->m_childMap.erase(childConnector);
	}

	hp<Connection>* phpConnection;
	if(phpConnection=m_parentConnections.searchForElement(
			context,sp<DAGNode>(this)))
	{
		const String parentConnector=(*phpConnection)->parentConnector();
		spNode->m_parentConnections.remove(sp<DAGNode>(this));
		spNode->m_parentMap.erase(parentConnector);
	}

	FE_DAG_VALIDATE();
	return TRUE;
}

void DAGNode::detach(void)
{
	FE_DAG_VALIDATE();
	List< hp<Connection> >::Iterator iterator(m_parentConnections);
	hp<Connection> hpConnection;
	while((hpConnection= *iterator++).isValid())
	{
		detachFrom(hpConnection->parent());
	}
	m_parentConnections.removeAll();
	m_parentMap.clear();
	FE_DAG_VALIDATE();
}

BWORD DAGNode::detach(String a_localConnector)
{
	std::map< String, hp<Connection> >::const_iterator it=
			m_parentMap.find(a_localConnector);
	if(it==m_parentMap.end())
	{
		return FALSE;
	}
	sp<Connection> spConnection=it->second;
	FEASSERT(spConnection.isValid());

	spConnection->parent()->m_childConnections.remove(spConnection);
	spConnection->parent()->m_childMap.erase(spConnection->parentConnector());

	m_parentConnections.remove(spConnection);
	m_parentMap.erase(a_localConnector);

	return TRUE;
}

sp<DAGNode::Connection> DAGNode::parentConnection(String a_parentConnector) const
{
	std::map< String, hp<Connection> >::const_iterator it=
			m_parentMap.find(a_parentConnector);
	if(it==m_parentMap.end())
	{
		return sp<Connection>(NULL);
	}
	return it->second;
}

sp<DAGNode::Connection> DAGNode::childConnection(String a_childConnector) const
{
	std::map< String, sp<Connection> >::const_iterator it=
			m_childMap.find(a_childConnector);
	if(it==m_childMap.end())
	{
		return sp<Connection>(NULL);
	}
	return it->second;
}


} // namespace
