/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __maya_SurfaceAccessorMaya_h__
#define __maya_SurfaceAccessorMaya_h__

#define FE_SAM_BLIND_DEBUG	FALSE

namespace fe
{

/**************************************************************************//**
    @brief Maya Surface Editing Implementation

	@ingroup maya

	Currently, this only supports poly meshes.
	We need to add NURBS support here or in a different class.
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessorMaya:
	public SurfaceAccessorBase
{
	public:

		enum	PolyFormat
				{
					e_unknown,
					e_closedPoly,
					e_openPoly,
					e_point
				};
						SurfaceAccessorMaya(void):
							m_pStringCache(NULL),
							m_pIntegerCache(NULL),
							m_pRealCache(NULL),
							m_pSpatialVectorCache(NULL),
							m_polyFormatID(-1),
							m_pFnMesh(NULL)
						{	setName("SurfaceAccessorMaya"); }
virtual					~SurfaceAccessorMaya(void)
						{
							if(m_pFnMesh)
							{
								delete m_pFnMesh;
							}
							if(m_pSpatialVectorCache)
							{
								delete[] m_pSpatialVectorCache;
							}
							if(m_pRealCache)
							{
								delete[] m_pRealCache;
							}
							if(m_pIntegerCache)
							{
								delete[] m_pIntegerCache;
							}
							if(m_pStringCache)
							{
								delete[] m_pStringCache;
							}
						}

		BWORD			bind(SurfaceAccessibleI::Element a_element,
								SurfaceAccessibleI::Attribute a_attribute)
						{
							FEASSERT(m_pFnMesh);

							m_attribute=a_attribute;
							if(a_attribute==SurfaceAccessibleI::e_vertices ||
									a_attribute==SurfaceAccessibleI::e_open)
							{
								FEASSERT(a_element==
										SurfaceAccessibleI::e_primitive);
								m_element=a_element;
								return TRUE;
							}

							return bindInternal(a_element,String());
						}
		BWORD			bind(SurfaceAccessibleI::Element a_element,
								const String& a_name)
						{
							FEASSERT(m_pFnMesh);

							m_attribute=SurfaceAccessibleI::e_generic;

							if(a_element==SurfaceAccessibleI::e_point)
							{
								if(a_name=="P")
								{
									m_attribute=SurfaceAccessibleI::e_position;
								}
								else if(a_name=="N")
								{
									m_attribute=SurfaceAccessibleI::e_normal;
								}
							}
							return bindInternal(a_element,a_name);
						}

						//* as SurfaceAccessorI

virtual	String			type(void) const	{ return "TODO"; }
virtual	U32				count(void) const
						{
							FEASSERT(m_pFnMesh);

							switch(m_element)
							{
							case SurfaceAccessibleI::e_point:
								return vertexCount();

//							case SurfaceAccessibleI::e_vertex:
//								return 0;

							case SurfaceAccessibleI::e_primitive:
								return m_pFnMesh->numPolygons();

//							case SurfaceAccessibleI::e_detail:
//								return 0;

							case SurfaceAccessibleI::e_pointGroup:
							case SurfaceAccessibleI::e_primitiveGroup:
								return m_groupList.size();

							default:
								;
							}
							return 0;
						}

virtual	U32				subCount(U32 a_index) const
						{
							FEASSERT(m_pFnMesh);

							if(m_attribute==SurfaceAccessibleI::e_vertices)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitive);

								return polyVertCount(a_index);
							}

							return 1;
						}

virtual	void			set(U32 a_index,U32 a_subIndex,String a_string)
						{
							FEASSERT(m_pFnMesh);

							assureBlindDataID("String");

							const I32 compID=rawVertex(a_index);
							const MFn::Type component=componentType();
							MStatus status=m_pFnMesh->setStringBlindData(
									compID,
									component,
									m_blindDataID,
									m_attrName.c_str(),
									a_string.c_str());
							if(!status)
							{
								feLog("SurfaceAccessorMaya::set(... String)"
										" index %d raw %d"
										" setStringBlindData failed\n",
										a_index, compID);
							}

							if(m_pStringCache)
							{
								m_pStringCache[a_index]=a_string;
							}
						}
virtual	String			string(U32 a_index,U32 a_subIndex=0)
						{
							FEASSERT(m_pFnMesh);

							if(m_pStringCache)
							{
								return m_pStringCache[a_index];
							}

							checkBlindDataID("I32");

							if(!m_pFnMesh->isBlindDataTypeUsed(m_blindDataID))
							{
#if FE_SAM_BLIND_DEBUG
								feLog("SurfaceAccessorMaya::integer"
										" blind data %p not used\n",
										m_blindDataID);
#endif
								return 0;
							}
							const MFn::Type component=componentType();
							if(!m_pFnMesh->hasBlindData(component))
							{
#if FE_SAM_BLIND_DEBUG
								feLog("SurfaceAccessorMaya::integer"
										" no component blind data\n");
#endif
								return 0;
							}
							const I32 compID=rawVertex(a_index);
							if(!m_pFnMesh->hasBlindDataComponentId(compID,
									component))
							{
#if FE_SAM_BLIND_DEBUG
								feLog("SurfaceAccessorMaya::integer"
										" no blind data on"
										" index %d compID %d\n",a_index,compID);
#endif
								return 0;
							}

							const I32 size=count();

							MIntArray compIDs(size);
							for(I32 m=0;m<size;m++)
							{
								compIDs.set(m,rawVertex(m));
							}

							MStringArray data;
							MStatus status=m_pFnMesh->getStringBlindData(
									component,
									m_blindDataID,
									m_attrName.c_str(),
									compIDs,
									data);
							if(!status)
							{
								feLog("SurfaceAccessorMaya::integer"
										" getStringBlindData failed\n");
								return 0;
							}

							//* TODO fix reverse lookup for curves

							m_pStringCache=new String[size];
							for(I32 m=0;m<size;m++)
							{
								m_pStringCache[compIDs[m]]=data[m].asChar();
							}

							return m_pStringCache[a_index];
						}

virtual	void			set(U32 a_index,U32 a_subIndex,I32 a_integer)
						{
							FEASSERT(m_pFnMesh);

							assureBlindDataID("I32");

							const I32 compID=rawVertex(a_index);
							const I32 data=a_integer;
							const MFn::Type component=componentType();
							MStatus status=m_pFnMesh->setIntBlindData(
									compID,
									component,
									m_blindDataID,
									m_attrName.c_str(),
									data);
							if(!status)
							{
								feLog("SurfaceAccessorMaya::set(... I32)"
										" index %d raw %d"
										" setIntBlindData failed\n",
										a_index, compID);
							}
							if(m_attrName=="POLY_FORMAT")
							{
								m_cachePolyFormat=PolyFormat(a_integer);
							}

							if(m_pIntegerCache)
							{
								m_pIntegerCache[a_index]=a_integer;
							}
						}
virtual	I32				integer(U32 a_index,U32 a_subIndex=0)
						{
							FEASSERT(m_pFnMesh);

							if(m_pIntegerCache)
							{
								return m_pIntegerCache[a_index];
							}

							checkBlindDataID("I32");

							if(m_element==SurfaceAccessibleI::e_pointGroup ||
									m_element==
									SurfaceAccessibleI::e_primitiveGroup)
							{
								return (a_index>=m_groupList.size())?
										-1: m_groupList[a_index];
							}

							if(m_attribute==SurfaceAccessibleI::e_open)
							{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitive ||
										m_element==
										SurfaceAccessibleI::e_primitiveGroup);

								const PolyFormat format=polyFormat(a_index);
								return format==e_openPoly;
							}

							if(!m_pFnMesh->isBlindDataTypeUsed(m_blindDataID))
							{
#if FE_SAM_BLIND_DEBUG
								feLog("SurfaceAccessorMaya::integer"
										" blind data %p not used\n",
										m_blindDataID);
#endif
								return 0;
							}
							const MFn::Type component=componentType();
							if(!m_pFnMesh->hasBlindData(component))
							{
#if FE_SAM_BLIND_DEBUG
								feLog("SurfaceAccessorMaya::integer"
										" no component blind data\n");
#endif
								return 0;
							}
							const I32 compID=rawVertex(a_index);
							if(!m_pFnMesh->hasBlindDataComponentId(compID,
									component))
							{
#if FE_SAM_BLIND_DEBUG
								feLog("SurfaceAccessorMaya::integer"
										" no blind data on"
										" index %d compID %d\n",a_index,compID);
#endif
								return 0;
							}

							const I32 size=count();

							MIntArray compIDs(size);
							for(I32 m=0;m<size;m++)
							{
								compIDs.set(m,rawVertex(m));
							}

							MIntArray data;
							MStatus status=m_pFnMesh->getIntBlindData(
									component,
									m_blindDataID,
									m_attrName.c_str(),
									compIDs,
									data);
							if(!status)
							{
								feLog("SurfaceAccessorMaya::integer"
										" getIntBlindData failed\n");
								return 0;
							}

							//* TODO fix reverse lookup for curves

							m_pIntegerCache=new I32[size];
							for(I32 m=0;m<size;m++)
							{
								m_pIntegerCache[compIDs[m]]=data[m];
							}

							return m_pIntegerCache[a_index];
						}
virtual	I32				duplicate(U32 index,U32 subIndex=0)
						{
							//* TODO
							return -1;
						}
virtual	void			append(U32 a_index,I32 a_integer)
						{
							if(m_element==SurfaceAccessibleI::e_pointGroup ||
									m_element==
									SurfaceAccessibleI::e_primitiveGroup)
							{
								// a_index ignored

								set(a_integer,U32(0),I32(true));

								for(I32 m=0;m<m_groupList.size();m++)
								{
									if(m_groupList[m]==a_integer)
									{
										return;
									}
								}
								m_groupList.push_back(a_integer);
							}
						}
virtual	void			remove(U32 a_index,I32 a_integer)
						{
							if(m_element==SurfaceAccessibleI::e_pointGroup ||
									m_element==
									SurfaceAccessibleI::e_primitiveGroup)
							{
								// a_index ignored

								set(a_integer,U32(0),I32(false));
								for(I32 m=0;m<m_groupList.size();m++)
								{
									if(m_groupList[m]==a_integer)
									{
										m_groupList.erase(
												m_groupList.begin()+m);
									}
								}
							}
						}

virtual	void			set(U32 a_index,U32 a_subIndex,Real a_real)
						{
							FEASSERT(m_pFnMesh);

							assureBlindDataID("Real");

							const I32 compID=rawVertex(a_index);

							const double data=a_real;
							const MFn::Type component=componentType();
							MStatus status=m_pFnMesh->setDoubleBlindData(
									compID,
									component,
									m_blindDataID,
									m_attrName.c_str(),
									data);
							if(!status)
							{
								feLog("SurfaceAccessorMaya::set(... Real)"
										" index %d compID %d"
										" setDoubleBlindData failed:"
										" \"%s\"\n",
										a_index, compID,
										status.errorString().asChar());
							}

							if(m_pRealCache)
							{
								m_pRealCache[a_index]=a_real;
							}
						}
virtual	Real			real(U32 a_index,U32 a_subIndex=0)
						{
							FEASSERT(m_pFnMesh);

							if(m_pRealCache)
							{
								return m_pRealCache[a_index];
							}

							checkBlindDataID("Real");

							if(!m_pFnMesh->isBlindDataTypeUsed(m_blindDataID))
							{
#if FE_SAM_BLIND_DEBUG
								feLog("SurfaceAccessorMaya::real"
										" blind data %p not used\n",
										m_blindDataID);
#endif
								return 0.0;
							}

							const MFn::Type component=componentType();
							if(!m_pFnMesh->hasBlindData(component))
							{
#if FE_SAM_BLIND_DEBUG
								feLog("SurfaceAccessorMaya::real"
										" no component blind data\n");
#endif
								return 0.0;
							}

							const I32 compID=rawVertex(a_index);
							if(!m_pFnMesh->hasBlindDataComponentId(compID,
									component))
							{
#if FE_SAM_BLIND_DEBUG
								feLog("SurfaceAccessorMaya::real"
										" no blind data on"
										" index %d compID %d\n",a_index,compID);
#endif
								return 0.0;
							}

							const I32 size=count();

							MIntArray compIDs(size);
							for(I32 m=0;m<size;m++)
							{
								compIDs.set(m,rawVertex(m));
							}

							MDoubleArray data;
							MStatus status=m_pFnMesh->getDoubleBlindData(
									component,
									m_blindDataID,
									m_attrName.c_str(),
									compIDs,
									data);
							if(!status)
							{
								feLog("SurfaceAccessorMaya::real"
										" getDoubleBlindData failed\n");
								return 0.0;
							}

							//* TODO fix reverse lookup for curves

							m_pRealCache=new Real[size];
							for(I32 m=0;m<size;m++)
							{
								m_pRealCache[compIDs[m]]=data[m];
							}

							return m_pRealCache[a_index];
						}

virtual	void			set(U32 a_index,U32 a_subIndex,
								const SpatialVector& a_vector)
						{
							FEASSERT(m_pFnMesh);

							MPoint point(a_vector[0],a_vector[1],a_vector[2]);

							if(m_element==SurfaceAccessibleI::e_pointGroup)
							{
								if(a_index>=m_groupList.size())
								{
									return;
								}
								a_index=m_groupList[a_index];
							}
							else if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								if(m_element==
										SurfaceAccessibleI::e_primitiveGroup)
								{
									if(a_index>=m_groupList.size())
									{
										return;
									}
									a_index=m_groupList[a_index];
								}
								else
								{
									FEASSERT(m_element==
											SurfaceAccessibleI::e_primitive);
								}

								U32 subs=m_pFnMesh->polygonVertexCount(a_index);
								FEASSERT(a_subIndex<subs);
								if(a_subIndex>=subs)
								{
									return;
								}

								MIntArray vertexList;
								MStatus status=m_pFnMesh->getPolygonVertices(
										a_index,vertexList);

								const U32 raw=vertexList[a_subIndex];
								status=m_pFnMesh->setPoint(raw,point);

								const PolyFormat format=polyFormat(a_index);
								if(format==e_openPoly)
								{
									const U32 raw2=
											vertexList[subs-1-a_subIndex];
									status=m_pFnMesh->setPoint(raw2,point);
								}
								return;
							}

							// TODO set color

							if(m_attribute==SurfaceAccessibleI::e_position ||
									m_element==SurfaceAccessibleI::e_pointGroup)
							{
								const U32 raw=rawVertex(a_index);
								MStatus status=m_pFnMesh->setPoint(raw,point);
								const PolyFormat format=polyFormat(0);	// HACK
								if(format==e_openPoly)
								{
									const U32 twin=twinVertex(a_index);
									status=m_pFnMesh->setPoint(twin,point);
								}
								return;
							}

							if(m_attribute==SurfaceAccessibleI::e_normal)
							{
								MVector vector(a_vector[0],a_vector[1],
										a_vector[2]);

								const U32 raw=rawVertex(a_index);
								MStatus status=
										m_pFnMesh->setVertexNormal(vector,raw);
								return;
							}

							assureBlindDataID("SpatialVector");

							I32 compID=rawVertex(a_index);
							for(U32 n=0;n<3;n++)
							{
								const double data=point[n];
								String name;
								name.sPrintf("%s_%c",m_attrName.c_str(),'x'+n);

								const MFn::Type component=componentType();
								MStatus status=m_pFnMesh->setDoubleBlindData(
										compID,
										component,
										m_blindDataID,
										name.c_str(),
										data);
								if(!status)
								{
									feLog("SurfaceAccessorMaya::set"
											"(... SpatialVector)"
											" index %d %d compID %d"
											" setDoubleBlindData failed:"
											" \"%s\"\n",
											a_index, n, compID,
											status.errorString().asChar());
								}
							}

							if(m_pSpatialVectorCache)
							{
								m_pSpatialVectorCache[a_index]=a_vector;
							}
						}
virtual	SpatialVector	spatialVector(U32 a_index,U32 a_subIndex=0)
						{
							FEASSERT(m_pFnMesh);

							if(m_pSpatialVectorCache)
							{
								return m_pSpatialVectorCache[a_index];
							}

							MPoint point;

							if(m_element==SurfaceAccessibleI::e_pointGroup)
							{
								if(a_index>=m_groupList.size())
								{
									return SpatialVector(0.0,0.0,0.0);
								}
								const U32 raw=rawVertex(m_groupList[a_index]);
								MStatus status=m_pFnMesh->getPoint(raw,point);
							}
							else if(m_element==
									SurfaceAccessibleI::e_primitiveGroup ||
									m_attribute==SurfaceAccessibleI::e_vertices)
							{
								if(m_element==
										SurfaceAccessibleI::e_primitiveGroup)
								{
									if(a_index>=m_groupList.size())
									{
										return SpatialVector(0.0,0.0,0.0);
									}
									a_index=m_groupList[a_index];
								}
								else
								{
								FEASSERT(m_element==
										SurfaceAccessibleI::e_primitive);
								}

								U32 subs=m_pFnMesh->polygonVertexCount(a_index);
								FEASSERT(a_subIndex<subs);
								if(a_subIndex>=subs)
								{
									return SpatialVector(0.0,0.0,0.0);
								}

								MIntArray vertexList;
								MStatus status=m_pFnMesh->getPolygonVertices(
										a_index,vertexList);

								const U32 raw=vertexList[a_subIndex];
								status=m_pFnMesh->getPoint(raw,point);
							}
							else if(m_attribute==SurfaceAccessibleI::e_position)
							{
								const U32 raw=rawVertex(a_index);
								MStatus status=m_pFnMesh->getPoint(raw,point);
							}
							else if(m_attribute==SurfaceAccessibleI::e_normal)
							{
								MVector vector;
								const U32 raw=rawVertex(a_index);
								MStatus status=
										m_pFnMesh->getVertexNormal(raw,vector);
								return SpatialVector(vector[0],
										vector[1],vector[2]);
							}
							else
							{
								// TODO get color

								checkBlindDataID("SpatialVector");

								if(!m_pFnMesh->isBlindDataTypeUsed(
										m_blindDataID))
								{
#if FE_SAM_BLIND_DEBUG
									feLog("SurfaceAccessorMaya::spatialVector"
											" blind data %p not used\n",
											m_blindDataID);
#endif
									return SpatialVector(0.0,0.0,0.0);
								}

								const MFn::Type component=componentType();
								if(!m_pFnMesh->hasBlindData(component))
								{
#if FE_SAM_BLIND_DEBUG
									feLog("SurfaceAccessorMaya::spatialVector"
											" no component blind data\n");
#endif
									return SpatialVector(0.0,0.0,0.0);
								}

								if(!m_pFnMesh->hasBlindDataComponentId(
										rawVertex(a_index),component))
								{
#if FE_SAM_BLIND_DEBUG
									feLog("SurfaceAccessorMaya::spatialVector"
											" no blind data on"
											" index %d (%d)\n",
											a_index,rawVertex(a_index));
#endif
									return SpatialVector(0.0,0.0,0.0);
								}

								const I32 size=count();
								m_pSpatialVectorCache=new SpatialVector[size];

								for(U32 n=0;n<3;n++)
								{
									String name;
									name.sPrintf("%s_%c",
											m_attrName.c_str(),'x'+n);

									MIntArray compIDs;
									MDoubleArray data;
									MStatus status=
											m_pFnMesh->getDoubleBlindData(
											component,
											m_blindDataID,
											name.c_str(),
											compIDs,
											data);
									if(!status)
									{
										feLog("SurfaceAccessorMaya::spatialVector"
												" getDoubleBlindData failed\n");

										delete m_pSpatialVectorCache;
										m_pSpatialVectorCache=NULL;

										return SpatialVector(0.0,0.0,0.0);
									}


									//* TODO fix reverse lookup for curves

									for(I32 m=0;m<size;m++)
									{
										m_pSpatialVectorCache[compIDs[m]][n]=
												data[m];
									}
								}

								return m_pSpatialVectorCache[a_index];
							}

							return SpatialVector(point[0],point[1],point[2]);
						}

						//* Maya specific
		void			setMeshObject(const MObject a_meshObject)
						{
							m_meshObject=a_meshObject;
							if(m_pFnMesh)
							{
								delete m_pFnMesh;
							}
							m_pFnMesh=new MFnMesh(a_meshObject);
							m_cachePolyFormat=PolyFormat(e_unknown);
							resetMark();
						}
const	MObject			meshObject(void)
						{	return m_meshObject; }

						//* TODO move to a shared base class
		void			setMaster(sp<Master> a_spMaster)
						{	m_spMaster=a_spMaster; }

	private:
		BWORD			bindInternal(SurfaceAccessibleI::Element a_element,
								const String& a_name)
						{
							m_element=SurfaceAccessibleI::e_point;
							if(a_element>=0 && a_element<=5)
							{
								m_element=a_element;
							}
							m_blindDataID= -1;
							m_attrName=a_name;
							updateGroupList();

							return TRUE;	//* TODO only TRUE if exists
						}

		void			updateGroupList(void);

		void			checkBlindDataID(const String& a_attrType)
						{
							if(m_attribute==SurfaceAccessibleI::e_generic &&
									m_blindDataID<0)
							{
								m_attrType=a_attrType;
								if(!m_attrName.empty())
								{
									const String attrKey=
											m_attrName+":"+m_attrType;
									m_blindDataID=
											m_spMaster->catalog()->catalog<I32>(
											"MayaBlindID",attrKey,-1);
								}
							}
						}
		void			assureBlindDataID(const String& a_attrType)
						{
							checkBlindDataID(a_attrType);
							if(m_attribute==SurfaceAccessibleI::e_generic &&
									m_blindDataID<0)
							{
								m_blindDataID=createBlindDataID(
										m_attrName,m_attrType);
							}
						}
		I32				createBlindDataID(String a_attrName,String a_attrType);

		void			checkPolyFormatID(void)
						{
							if(m_polyFormatID<0)
							{
								const String attrKey="POLY_FORMAT:I32";
								m_polyFormatID=
										m_spMaster->catalog()->catalog<I32>(
										"MayaBlindID",attrKey,-1);
							}
						}
		void			assurePolyFormatID(void)
						{
							checkPolyFormatID();
							if(m_polyFormatID<0)
							{
								m_polyFormatID=createBlindDataID(
										"POLY_FORMAT","I32");
							}

							//* This is apparently needed before hasBlindData()
							//* to wake it up or something.
							//* It doesn't have to be the same one we check.
							I32 probeInt= -1;
							m_pFnMesh->getIntBlindData(0,
									MFn::kMeshPolygonComponent,
									m_polyFormatID, "POLY_FORMAT",
									probeInt);
						}
		PolyFormat		polyFormat(U32 a_index) const
						{
							//* TODO cache all polys, like with m_pIntegerCache

							if(m_cachePolyFormat!=e_unknown)
							{
								return m_cachePolyFormat;
							}
							const_cast<SurfaceAccessorMaya*>(this)
									->assurePolyFormatID();

							I32 data=0;
							MStatus status=m_pFnMesh->getIntBlindData(
									a_index,
									MFn::kMeshPolygonComponent,
									m_polyFormatID,
									"POLY_FORMAT",
									data);
							if(!status)
							{
								feLog("SurfaceAccessorMaya::polyFormat"
										" getIntegerBlindData failed\n");
								return e_closedPoly;
							}
							const_cast<SurfaceAccessorMaya*>(this)->
									m_cachePolyFormat=PolyFormat(data);
							return m_cachePolyFormat;
						}

		U32				polyVertCount(U32 a_index) const
						{
							FEASSERT(m_pFnMesh);

							const U32 rawCount=
									m_pFnMesh->polygonVertexCount(a_index);
							const PolyFormat format=polyFormat(a_index);
							switch(format)
							{
								case e_closedPoly:
									return rawCount;
								case e_openPoly:
									return rawCount>>1;
								case e_point:
									return rawCount? 1: 0;
								default:
									;
							}

							return 0;
						}
		U32				vertexCount(void) const
						{
							const_cast<SurfaceAccessorMaya*>(this)
									->assurePolyFormatID();

							if(!m_pFnMesh->hasBlindData(
									MFn::kMeshPolygonComponent,m_polyFormatID))
							{
								return m_pFnMesh->numVertices();
							}
							//* TODO account for potentially shared points
							return countVertices();
						}
		U32				countVertices(void) const
						{
							// TODO cache
							U32 sum=0;
							U32 primitiveCount=m_pFnMesh->numPolygons();
							for(U32 index=0;index<primitiveCount;index++)
							{
								sum+=polyVertCount(index);
							}
							return sum;
						}
		void			resetMark(void)
						{
							m_markPoly=0;
							m_markVertex=0;
							m_markRawVertex=0;
						}
		U32				rawVertex(U32 a_logicalVertex)
						{
							assurePolyFormatID();
							if(!m_pFnMesh->hasBlindData(
									MFn::kMeshPolygonComponent,m_polyFormatID))
							{
								return a_logicalVertex;
							}
							while(TRUE)
							{
								if(a_logicalVertex<m_markVertex)
								{
									if(m_markPoly<1)
									{
										feX("SurfaceAccessorMaya::rawVertex",
												"mark data corrupt");
									}
									m_markPoly--;
									m_markVertex-=polyVertCount(m_markPoly);
									m_markRawVertex-=
											m_pFnMesh->polygonVertexCount(
											m_markPoly);
									continue;
								}
								const U32 subIndex=a_logicalVertex-m_markVertex;
								const U32 subs=polyVertCount(m_markPoly);
								if(subIndex<subs)
								{
									return m_markRawVertex+subIndex;
								}
								const U32 primitiveCount=
										m_pFnMesh->numPolygons();
								if(m_markPoly>=primitiveCount-1)
								{
									feX("SurfaceAccessorMaya::rawVertex",
											"excessive point index");
								}
								m_markVertex+=polyVertCount(m_markPoly);
								m_markRawVertex+=
										m_pFnMesh->polygonVertexCount(
										m_markPoly);
								m_markPoly++;
							}
						}
		U32				twinVertex(U32 a_logicalVertex)
						{
							rawVertex(a_logicalVertex);
							const U32 subIndex=a_logicalVertex-m_markVertex;
							const U32 subs=m_pFnMesh->polygonVertexCount(
											m_markPoly);

							MIntArray vertexList;
							MStatus status=m_pFnMesh->getPolygonVertices(
									m_markPoly,vertexList);

							return vertexList[subs-1-subIndex];
						}

		MFn::Type		componentType(void)
						{
							switch (m_element)
							{
								case SurfaceAccessibleI::e_point:
								case SurfaceAccessibleI::e_pointGroup:
									return MFn::kMeshVertComponent;
								case SurfaceAccessibleI::e_primitive:
								case SurfaceAccessibleI::e_primitiveGroup:
									return MFn::kMeshPolygonComponent;
								default:
									;
							}
							return MFn::kInvalid;
						}

										//* cache last raw vertex lookup
		U32								m_markPoly;
		U32								m_markVertex;
		U32								m_markRawVertex;

		PolyFormat						m_cachePolyFormat;	// TODO mixed

		sp<Master>						m_spMaster;
		MObject							m_meshObject;
		MFnMesh*						m_pFnMesh;
		String							m_attrType;
		I32								m_blindDataID;
		I32								m_polyFormatID;

		std::vector<I32>				m_groupList;

		String*							m_pStringCache;
		I32*							m_pIntegerCache;
		Real*							m_pRealCache;
		SpatialVector*					m_pSpatialVectorCache;
};

} // namespace

#endif /* __maya_SurfaceAccessorMaya_h__ */
