/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __core_Catalog_h__
#define __core_Catalog_h__

#define FE_CAT_SPEC(name, description)	name
#define FE_CAT(name)					name

namespace fe
{

/**************************************************************************//**
	@brief Dictionary of Arbitrary Instances

	@ingroup plugin

	TODO allow reordering
	TODO allow individual removal (remember to also remove from order map)
*//***************************************************************************/
class FE_DL_EXPORT Catalog: virtual public Component
{
	public:
							Catalog(void);
virtual						~Catalog(void);

							/** @brief Returns the number of name keys

								Each name key may have any number of properties.
							*/
		U32					catalogSize(void);

							/** @brief Returns a key by ordered index

								By default, keys are index in creation order.
							*/
		String				catalogKey(U32 index);

							/** @brief Returns keys in order of creation

								Keys are appended to the given array.
								Clear the array if only want new values.

								Future improvements may allow reordering.
							*/
		void				catalogKeys(std::vector<String> &a_keys);

							/** @brief Returns TRUE if the given property for
								the given name has been created.

								The default property if 'value'.
							*/
		BWORD				cataloged(String a_name,
									String a_property="value") const;

							/** @brief Returns TRUE any property for
								the given name has been created.
							*/
		BWORD				catalogedAny(String a_name) const;

							/** @brief Remove all entries.
							*/
		void				catalogClear(void);

							/** @brief Returns a component entry of the given
								property for the given name.

								If the entry does not exist,
								it may be created using the given
								implementation name.

								The return value will be invalid if the
								implementation could not be instantiated.
							*/
		sp<Component>		catalogComponent(String implementation,
									String a_name,String a_property="value");

							/** @brief Returns a component entry of the given
								property for the given name.

								Regardless of whether the entry previously
								existed, it is replaced by the given Component.

								The default property if 'value'.
							*/
		sp<Component>		catalogComponent(sp<Component> a_spComponent,
									String a_name,String a_property="value");

							/** @brief Returns the literal Instance of the
								given property for the given name.

								Regardless of whether the entry previously
								existed, it is replaced by the given Instance.

								The default property if 'value'.
							*/
		Instance&			catalogInstance(Instance &instance,
									String a_name,String a_property="value");

							/** @brief Returns the literal Instance of the
								given property for the given name.

								If the entry does not exist, it will be created.

								The default property if 'value'.
							*/
		Instance&			catalogInstance(String a_name,
									String a_property="value");

							/** @brief Returns existing catalog entry
								if already set or a default if not

								If entry does not exist, this version will
								create an entry and set it to the appropriate
								Type default.

								The default property if 'value'.
							*/
		template <class T>
		T&					catalog(String a_name,String a_property="value");

							/** @brief Returns existing catalog entry
								if already set or default if not

								If entry does not exist, this version will
								create an entry and set it to the given default.

								The default property if 'value'.
							*/
		template <class T>
		T&					catalog(String a_name,String a_property,
									const T& a_default);
//		template <class T>
//		T&					catalog(String a_name,const T& a_default)
//							{	return catalog(a_name,"value",a_default); }

							/** @brief Returns existing catalog entry
								if already set or default if not

								If entry does not exist, this version will
								not create an entry and simply return
								the given default.

								The default property if 'value'.
							*/
		template <class T>
const	T&					catalogOrDefault(String a_name,String a_property,
									const T& a_default);
		template <class T>
const	T&					catalogOrDefault(String a_name,const T& a_default)
							{	return catalogOrDefault(a_name,
										"value",a_default); }

		bool				catalogLookup(const String a_name,
									String a_property,
									Instance &a_instance);
		bool				catalogLookup(const String a_name,
									Instance &a_instance)
							{	return catalogLookup(a_name,"value",
										a_instance); }

							///	@brief Move all properties for name to bottom.
		void				catalogMoveToEnd(String a_name);

							///	@brief Remove all properties for a name.
		void				catalogRemove(String a_name);

							/** @brief Print a description of the catalog
								to the log.
							*/
		void				catalogDump(void);

	private:

		sp<TypeMaster>		typeMaster(void) const;

		template<class T>
		T&					createInstance(String a_name,
									String a_property="value");

		std::map< String, InstanceMap >	m_instanceMapMap;
		std::map< I32, String >			m_orderMap;
		I32								m_highest;
};

template<class T>
T& Catalog::createInstance(String a_name,String a_property)
{
	if(!catalogedAny(a_name))
	{
		m_orderMap[++m_highest]=a_name;
	}
	T& rT=m_instanceMapMap[a_name][a_property].create<T>(typeMaster());
#if FE_COUNTED_TRACK
	Counted::trackReference(&rT,this,
			String("Catalog ")+typeid(T).name());
#endif
	return rT;
}

template<class T>
T &Catalog::catalog(String a_name,String a_property)
{
	if(!cataloged(a_name,a_property))
	{
		return createInstance<T>(a_name,a_property);
	}
	return m_instanceMapMap[a_name][a_property].cast<T>();
}

template<class T>
T &Catalog::catalog(String a_name,String a_property,const T& a_default)
{
	if(!cataloged(a_name,a_property))
	{
		createInstance<T>(a_name,a_property)=a_default;
	}
	return m_instanceMapMap[a_name][a_property].cast<T>();
}

template<class T>
const T& Catalog::catalogOrDefault(String a_name,String a_property,
	const T& a_default)
{
	if(!cataloged(a_name,a_property))
	{
		return a_default;
	}
	return m_instanceMapMap[a_name][a_property].cast<T>();
}

inline sp<Component> Catalog::catalogComponent(String implementation,
		String a_name,String a_property)
{
	sp<Component> spComponent;
	if(cataloged(a_name,a_property))
	{
		Instance &instance = m_instanceMapMap[a_name][a_property];
		if(instance.is< sp<Component> >())
		{
			spComponent = instance.cast< sp<Component> >();
		}
	}
	if(!spComponent.isValid())
	{
		spComponent=registry()->create(implementation,a_name);
		createInstance< sp<Component> >(a_name,a_property)=spComponent;
	}
	return spComponent;
}

inline sp<Component> Catalog::catalogComponent(sp<Component> a_spComponent,
		String a_name,String a_property)
{
	sp<Component>& rspComponent=
			createInstance< sp<Component> >(a_name,a_property);

	rspComponent=a_spComponent;

	return a_spComponent;
}

inline Instance &Catalog::catalogInstance(String a_name,String a_property)
{
	return m_instanceMapMap[a_name][a_property];
}

inline Instance &Catalog::catalogInstance(Instance &instance,String a_name,
		String a_property)
{
	m_instanceMapMap[a_name][a_property]=instance;
	return m_instanceMapMap[a_name][a_property];
}

inline BWORD Catalog::catalogedAny(String a_name) const
{
	std::map<String, InstanceMap>::const_iterator it=
			m_instanceMapMap.find(a_name);

	return it!=m_instanceMapMap.end();
}

inline BWORD Catalog::cataloged(String a_name,String a_property) const
{
	std::map<String, InstanceMap>::const_iterator it=
			m_instanceMapMap.find(a_name);

	return it!=m_instanceMapMap.end() &&
			it->second.find(a_property) != it->second.end();
}

inline void Catalog::catalogMoveToEnd(String a_name)
{
	for(std::map< I32, String >::iterator it = m_orderMap.begin();
			it != m_orderMap.end(); it++)
	{
		if(it->second==a_name)
		{
			m_orderMap.erase(it->first);
			m_orderMap[++m_highest]=a_name;
			return;
		}
	}
}

inline void Catalog::catalogRemove(String a_name)
{
#if FE_COUNTED_TRACK
	std::map<String, InstanceMap>::iterator i_name=
			m_instanceMapMap.find(a_name);

	if(i_name!=m_instanceMapMap.end())
	{
		InstanceMap& rInstanceMap=i_name->second;
		for(std::map< String, Instance >::iterator i_prop=rInstanceMap.begin();
				i_prop!=rInstanceMap.end(); i_prop++)
		{
			Counted::untrackReference(i_prop->second.data(),this);
		}
	}
#endif
	m_instanceMapMap.erase(a_name);

	for(std::map< I32, String >::iterator it = m_orderMap.begin();
			it != m_orderMap.end(); it++)
	{
		if(it->second==a_name)
		{
			m_orderMap.erase(it->first);
			break;
		}
	}
}

inline void Catalog::catalogClear(void)
{
#if FE_COUNTED_TRACK
	for(std::map< String, InstanceMap >::iterator i_name=
			m_instanceMapMap.begin(); i_name!=m_instanceMapMap.end(); i_name++)
	{
		InstanceMap& rInstanceMap=i_name->second;
		for(std::map< String, Instance >::iterator i_prop=rInstanceMap.begin();
				i_prop!=rInstanceMap.end(); i_prop++)
		{
			Counted::untrackReference(i_prop->second.data(),this);
		}
	}
#endif
	m_instanceMapMap.clear();
	m_orderMap.clear();
}

} /* namespace */

#endif /* __core_Catalog_h__ */
